"""Tests for the views of the aps_bom app."""
import os

from django.conf import settings
from django.core.files.uploadedfile import SimpleUploadedFile
from django.test import TestCase

from django_libs.tests.mixins import ViewTestMixin, ViewRequestFactoryTestMixin
from django_libs.tests.factories import UserFactory

from .. import views
from . import factories


class DropDownChoicesAJAXViewTestCase(ViewRequestFactoryTestMixin, TestCase):
    """Tests for the ``DropDownChoicesAJAXView`` view class."""
    view_class = views.DropDownChoicesAJAXView

    def setUp(self):
        self.user = UserFactory()
        self.company = factories.CompanyFactory()
        self.epn = factories.EPNFactory(company=self.company)
        self.data = {
            'action': 'get_epn_choices_for_customer',
            'obj_id': self.company.id,
        }

    def test_view(self):
        self.should_redirect_to_login_when_anonymous()
        self.is_not_callable(user=self.user)
        self.is_callable(user=self.user, ajax=True, data=self.data)


class BOMUploadViewTestCase(ViewTestMixin, TestCase):
    """Tests for the ``BOMUploadView`` view class."""
    longMessage = True

    def get_view_name(self):
        return 'aps_bom_bom_upload'

    def get_login_url(self):
        return settings.LOGIN_URL

    def setUp(self):
        self.user = UserFactory()
        self.unit = factories.UnitFactory(code='pcs', description='pieces')

        self.ipn = factories.IPNFactory()
        factories.IPNFactory(code='7900.1500')
        factories.IPNFactory(code='6030.2345')

        self.csv_file = open(os.path.join(
            settings.APP_ROOT, 'tests/files/BOM.csv'))

        self.data = {
            'ipn': self.ipn.pk,
            'description': 'Test description for BOM',
            'csv_file': SimpleUploadedFile('BOM.csv', self.csv_file.read())
        }

    def tearDown(self):
        self.csv_file.close()

    def test_view(self):
        self.should_redirect_to_login_when_anonymous()
        self.should_be_callable_when_authenticated(self.user)
        self.is_callable(method='post', data=self.data)


class CBOMDownloadViewTestCase(ViewTestMixin, TestCase):
    """Tests for the ``CBOMDownloadView`` view class."""
    longMessage = True

    def get_login_url(self):
        return settings.LOGIN_URL

    def get_view_kwargs(self):
        return {'cbom_pk': self.cbom.pk}

    def get_view_name(self):
        return 'aps_bom_cbom_download'

    def setUp(self):
        self.user = UserFactory()
        self.cbom = factories.CBOMFactory()

    def test_view(self):
        self.should_redirect_to_login_when_anonymous()
        self.should_be_callable_when_authenticated(self.user)
        self.is_not_callable(kwargs={'cbom_pk': 123}, message=(
            'With the wrong kwargs, the view should not be callable.'))
        self.is_callable(kwargs={}, message=(
            'Even without kwargs, the view should be callable.'))


class CBOMUploadViewTestCase(ViewTestMixin, TestCase):
    """Tests for the ``CBOMUploadView`` view class."""
    longMessage = True

    def get_view_name(self):
        return 'aps_bom_cbom_upload'

    def get_login_url(self):
        return settings.LOGIN_URL

    def setUp(self):
        self.user = UserFactory()
        self.customer = factories.CompanyFactory()
        self.unit = factories.UnitFactory(code='pcs', description='pieces')

        factories.EPNFactory(epn='945382', company=self.customer)
        factories.EPNFactory(epn='743629', company=self.customer)

        self.csv_file = open(os.path.join(
            settings.APP_ROOT, 'tests/files/cBOM.csv'))

        self.data = {
            'customer': self.customer.id,
            'description': 'Description for the CBOM',
            'html_link': '',
            'product': 'This product is used here',
            'version_date': '12/12/2012',
            'csv_file': SimpleUploadedFile('cBOM.csv', self.csv_file.read())
        }

    def tearDown(self):
        self.csv_file.close()

    def test_view(self):
        self.should_redirect_to_login_when_anonymous()
        self.should_be_callable_when_authenticated(self.user)
        self.is_callable(method='post', data=self.data)



"""Factories for the aps_bom app."""
from decimal import Decimal

from django.utils.timezone import now

import factory

from ..models import (
    AdditionalText,
    BOM,
    BOMItem,
    CBOM,
    CBOMItem,
    Company,
    Country,
    EPN,
    IPN,
    PriceGroup,
    PriceMarker,
    Shape,
    Unit,
)


class AdditionalTextFactory(factory.DjangoModelFactory):
    """Factory for the ``AdditionalText`` model."""
    FACTORY_FOR = AdditionalText

    ipn = factory.SubFactory('aps_bom.tests.factories.IPNFactory')
    text = factory.Sequence(lambda n: 'text {0}'.format(n))


class BOMFactory(factory.DjangoModelFactory):
    """Factory for the ``BOM`` model."""
    FACTORY_FOR = BOM

    description = factory.Sequence(lambda n: 'description {0}'.format(n))
    ipn = factory.SubFactory('aps_bom.tests.factories.IPNFactory')


class BOMItemFactory(factory.DjangoModelFactory):
    """Factory for the ``BOMItem`` model."""
    FACTORY_FOR = BOMItem

    bom = factory.SubFactory(BOMFactory)
    ipn = factory.SubFactory('aps_bom.tests.factories.IPNFactory')
    position = factory.Sequence(lambda n: '{0}'.format(n))
    qty = factory.Sequence(lambda n: n)
    unit = factory.SubFactory('aps_bom.tests.factories.UnitFactory')


class CBOMFactory(factory.DjangoModelFactory):
    """Factory for the ``CBOM`` model."""
    FACTORY_FOR = CBOM

    customer = factory.SubFactory('aps_bom.tests.factories.CompanyFactory')
    description = factory.Sequence(lambda n: 'description {0}'.format(n))
    html_link = factory.Sequence(lambda n: 'http://example.com/{0}/'.format(n))
    product = factory.Sequence(lambda n: 'product {0}'.format(n))
    version_date = now()


class CBOMItemFactory(factory.DjangoModelFactory):
    """Factory for the ``CBOMItem`` model."""
    FACTORY_FOR = CBOMItem

    bom = factory.SubFactory(CBOMFactory)
    epn = factory.SubFactory('aps_bom.tests.factories.EPNFactory')
    position = factory.Sequence(lambda n: '{0}'.format(n))
    qty = factory.Sequence(lambda n: n)
    unit = factory.SubFactory('aps_bom.tests.factories.UnitFactory')


class CompanyFactory(factory.DjangoModelFactory):
    """Factory for the ``Company`` model."""
    FACTORY_FOR = Company

    country = factory.SubFactory('aps_bom.tests.factories.CountryFactory')
    code = factory.Sequence(lambda n: 'code {0}'.format(n))
    description = factory.Sequence(lambda n: 'description {0}'.format(n))


class CountryFactory(factory.DjangoModelFactory):
    """Factory for the ``Country`` model."""
    FACTORY_FOR = Country

    code = 'de'
    description = factory.Sequence(lambda n: 'description {0}'.format(n))


class EPNFactory(factory.DjangoModelFactory):
    """Factory for the ``EPN`` model."""
    FACTORY_FOR = EPN

    company = factory.SubFactory(CompanyFactory)
    cpn = factory.SubFactory('aps_bom.tests.factories.IPNFactory')
    description = factory.Sequence(lambda n: 'description {0}'.format(n))
    epn = factory.Sequence(lambda n: 'epn {0}'.format(n))
    ipn = factory.SubFactory('aps_bom.tests.factories.IPNFactory')


class IPNFactory(factory.DjangoModelFactory):
    """Factory for the ``IPN`` model."""
    FACTORY_FOR = IPN

    code = factory.Sequence(lambda n: 'code {0}'.format(n))
    name = factory.Sequence(lambda n: 'name {0}'.format(n))
    price_group = factory.SubFactory(
        'aps_bom.tests.factories.PriceGroupFactory')
    shape = factory.SubFactory('aps_bom.tests.factories.ShapeFactory')
    price_max = Decimal('123.45')


class PriceGroupFactory(factory.DjangoModelFactory):
    """Factory for the ``PriceGroup`` model."""
    FACTORY_FOR = PriceGroup

    code = 'A'


class PriceMarkerFactory(factory.DjangoModelFactory):
    """Factory for the ``PriceMarker`` model."""
    FACTORY_FOR = PriceMarker

    ipn = factory.SubFactory(IPNFactory)
    price = Decimal('54.321')
    date = now()


class ShapeFactory(factory.DjangoModelFactory):
    """Factory for the ``Shape`` model."""
    FACTORY_FOR = Shape

    code = factory.Sequence(lambda n: 'code {0}'.format(n))
    name = factory.Sequence(lambda n: 'name {0}'.format(n))


class UnitFactory(factory.DjangoModelFactory):
    """Factory for the ``Unit`` model."""
    FACTORY_FOR = Unit

    code = factory.Sequence(lambda n: 'c{0}'.format(n))
    description = factory.Sequence(lambda n: 'description {0}'.format(n))



"""Models for the ``aps_bom`` app."""
import os
from csv import DictWriter

from decimal import Decimal

from django.conf import settings
from django.db import models
from django.forms.models import model_to_dict
from django.utils.translation import ugettext_lazy as _

from . import settings as app_settings


# TODO General note: attributes in docstrings, where it was not entirely clear,
# what they refer to are marked with TODO for further update.

def trunc(data, length=50):  # pragma: nocover
    """Shortcut for truncating strings."""
    return (data[:length] + '..') if len(data) > length else data


class AdditionalText(models.Model):
    """
    Additional text, that can be added to an IPN.

    :ipn: The IPN of what this text is about.
    :text: The actual text, that is attached.

    """
    ipn = models.ForeignKey(
        'aps_bom.IPN',
        verbose_name=_('IPN'),
        related_name='additional_texts',
    )

    text = models.CharField(
        verbose_name=_('Text'),
        max_length=100,
    )

    def __unicode__(self):
        return u'Text "{0}" for {1}'.format(trunc(self.text), self.ipn.code)


class BOMManager(models.Manager):
    def get_queryset(self):
        return super(BOMManager, self).get_queryset().select_related('ipn')


class BOM(models.Model):
    """
    Bill Of Materials.

    :description: The description of this bill.
    :ipn: The IPN of what is billed.

    reverse relations:
    :bomitems: All the BOMItems on this BOM.

    """
    description = models.CharField(
        verbose_name=_('Description'),
        max_length=128,
    )

    ipn = models.ForeignKey(
        'aps_bom.IPN',
        verbose_name=_('IPN'),
        related_name='boms',
        blank=True, null=True,
    )

    objects = BOMManager()

    def __unicode__(self):
        if self.ipn_id is not None:
            ipn_code = self.ipn.code
        else:
            ipn_code = None
        return u'BOM "{0}" for {1}'.format(trunc(self.description), ipn_code)

    class Meta:
        ordering = ('ipn__code', )


class BOMItemManager(models.Manager):
    def get_queryset(self):
        return super(BOMItemManager, self).get_queryset().select_related(
            'ipn', 'bom', 'unit')


class BOMItem(models.Model):
    """
    Details about an item, that belongs to a BOM.

    :bom: The BOM this item belongs to.
    :ipn: The IPN of this item.
    :position: The position on the list of items.
    :qty: The quantity of this item.
    :unit: The unit, that qty refers to.

    """
    bom = models.ForeignKey(
        'aps_bom.BOM',
        verbose_name=_('BOM'),
        related_name='bomitems',
    )

    ipn = models.ForeignKey(
        'aps_bom.IPN',
        verbose_name=_('IPN'),
        related_name='bomitems',
    )

    position = models.CharField(
        verbose_name=_('Position'),
        max_length=5,
    )

    qty = models.FloatField(
        verbose_name=_('Quantity'),
    )

    unit = models.ForeignKey(
        'aps_bom.Unit',
        verbose_name=_('Unit'),
        related_name='bomitems',
    )

    objects = BOMItemManager()

    def __unicode__(self):
        return u'{0} {1} of item "{2}" of BOM "{3}"'.format(
            self.qty, self.unit.code, self.ipn.code,
            trunc(self.bom.description))


class CBOMManager(models.Manager):
    def get_queryset(self):
        return super(CBOMManager, self).get_queryset().select_related(
            'customer')


class CBOM(models.Model):
    """
    Customer Bill Of Materials.

    :customer: The customer this bill belongs to.
    :description: A description of the product.
    :html_link: TODO
    :product: The name of the product.
    :version_date: The date of this version of the bill.

    reverse relations:
    :cbomitems: All the CBOMItems on this BOM

    """
    customer = models.ForeignKey(
        'aps_bom.Company',
        verbose_name=_('Customer'),
        related_name='cboms',
    )

    description = models.CharField(
        verbose_name=_('Description'),
        max_length=128,
    )

    html_link = models.URLField(
        verbose_name=_('HTML link'),
        max_length=256,
        blank=True,
    )

    product = models.CharField(
        verbose_name=_('Product'),
        max_length=100,
    )

    version_date = models.DateField(
        verbose_name=_('Version date'),
    )

    objects = CBOMManager()

    def __unicode__(self):
        return u'{0} from "{1}" ({2})'.format(
            self.product, self.customer.description, self.version_date)

    def get_bom(self):
        """Converts this CBOM into a matching BOM."""
        return BOM(description=self.description)

    def get_bom_csv_file(self):
        """
        Returns the url for a BOM.csv file created from this CBOM instance.

        """
        file_name = 'BOM_{0}.csv'.format(self.id)
        path = os.path.join(settings.MEDIA_ROOT, file_name)
        with open(path, 'w') as f:
            csv_writer = DictWriter(f, app_settings.BOM_CSV_FIELDNAMES)
            csv_writer.writeheader()
            clean_list = []
            # for each bomitem...
            for bomitem in self.get_bom_items():
                # ... create a dictionary
                bomitem_dict = model_to_dict(bomitem)
                # ... clean the values
                clean_dict = {}
                for fieldname in app_settings.BOM_CSV_FIELDNAMES:
                    if fieldname.lower() == 'description':
                        clean_dict.update({fieldname: bomitem.ipn.name})
                    elif fieldname.lower() == 'ipn':
                        clean_dict.update({fieldname: bomitem.ipn.code})
                    elif fieldname.lower() == 'shape':
                        clean_dict.update({fieldname: bomitem.ipn.shape.code})
                    elif fieldname.lower() == 'unit':
                        clean_dict.update({fieldname: bomitem.unit.code})
                    else:
                        clean_dict.update({
                            fieldname: bomitem_dict[fieldname.lower()]})
                # ... append the dict to the list
                clean_list.append(clean_dict)
            #... and write the entire list of dicts to the file
            csv_writer.writerows(clean_list)
        return os.path.join(settings.MEDIA_URL, file_name)

    def get_bom_items(self):
        """Converts all CBOMItems into matching BOMItems."""
        bomitems = []
        bom = self.get_bom()
        for cbomitem in self.cbomitems.all():
            if cbomitem.consign:
                ipn = cbomitem.epn.cpn
            else:
                ipn = cbomitem.epn.ipn
            bomitems.append(BOMItem(bom=bom, ipn=ipn,
                                    position=cbomitem.position,
                                    qty=cbomitem.qty, unit=cbomitem.unit))
        return bomitems

    def get_csv_file(self):
        """
        Returns the url for a cBOM.csv file created from this CBOM instance.

        """
        file_name = 'cBOM_{0}.csv'.format(self.id)
        path = os.path.join(settings.MEDIA_ROOT, file_name)
        with open(path, 'w') as f:
            csv_writer = DictWriter(f, app_settings.CBOM_CSV_FIELDNAMES)
            csv_writer.writeheader()
            clean_list = []
            # for each bomitem...
            for cbomitem in self.cbomitems.all():
                # ... create a dictionary
                cbomitem_dict = model_to_dict(cbomitem)
                # ... clean the values
                clean_dict = {}
                for fieldname in app_settings.CBOM_CSV_FIELDNAMES:
                    if fieldname.lower() == 'epn':
                        clean_dict.update({fieldname: cbomitem.epn.epn})
                    elif fieldname.lower() == 'consign':
                        clean_dict.update({fieldname: int(cbomitem.consign)})
                    elif fieldname.lower() == 'unit':
                        clean_dict.update({fieldname: cbomitem.unit.code})
                    elif fieldname.lower() == 'description':
                        clean_dict.update({
                            fieldname: cbomitem.epn.description})
                    else:
                        clean_dict.update({
                            fieldname: cbomitem_dict[fieldname.lower()]})
                # ... append the dict to the list
                clean_list.append(clean_dict)
            #... and write the entire list of dicts to the file
            csv_writer.writerows(clean_list)
        return os.path.join(settings.MEDIA_URL, file_name)


class CBOMItemManager(models.Manager):
    def get_queryset(self):
        return super(CBOMItemManager, self).get_queryset().select_related(
            'bom', 'epn', 'epn__company', 'unit')


class CBOMItem(models.Model):
    """
    Details about the items, that belong to a CBOM.

    :bom: The CBOM this item belongs to.
    :consign: Whether to use the IPN or CPN, when translating it into BOMItem.
      True means use CPN, False therefore uses the IPN of the EPN.
    :epn: The external part number of of this item.
    :position: The position of this item on th CBOM list of items.
    :qty: The quantity of this item on the list.
    :unit: The unit qty refers to.

    """

    bom = models.ForeignKey(
        'aps_bom.CBOM',
        verbose_name=_('CBOM'),
        related_name='cbomitems',
    )

    consign = models.BooleanField(
        verbose_name=_('Consign'),
        default=True,
    )

    epn = models.ForeignKey(
        'aps_bom.EPN',
        verbose_name=_('EPN'),
        related_name='cbomitems',
    )

    position = models.CharField(
        verbose_name=_('Position'),
        max_length=5,
    )

    qty = models.FloatField(
        verbose_name=_('Quantity'),
    )

    unit = models.ForeignKey(
        'aps_bom.Unit',
        verbose_name=_('Unit'),
        related_name='cbomitems',
    )

    # fields for caching the related values to increase performance
    unit_code = models.CharField(max_length=5, blank=True)
    epn_code = models.CharField(max_length=513, blank=True)
    bom_short_description = models.CharField(max_length=128, blank=True)

    objects = CBOMItemManager()

    def __unicode__(self):
        return u'{0} {1} of item "{2}" of CBOM "{3}"'.format(
            self.qty, self.unit_code, self.epn_code,
            self.bom_short_description)

    def save(self, *args, **kwargs):
        self.unit_code = self.unit.code
        self.epn_code = self.epn.__unicode__()
        self.bom_short_description = trunc(self.bom.description)
        return super(CBOMItem, self).save(*args, **kwargs)


class Company(models.Model):
    """
    Holds basic information about a company.

    :country: The country, this company is in.
    :code: A short identifier for this company.
    :description: Full company description (name).

    reverse relations:
    :cboms: All the CBOMs related to this company.
    :epns: All the EPNs from this country.

    """
    country = models.ForeignKey(
        'aps_bom.Country',
        verbose_name=_('Country'),
        related_name='companies',
    )

    code = models.CharField(
        verbose_name=_('Code'),
        max_length=10,
    )

    description = models.CharField(
        verbose_name=_('Description'),
        max_length=128,
    )

    def __unicode__(self):
        return u'Customer "{0}"'.format(self.description)


class Country(models.Model):
    """
    Information about a country.

    :code: Country code.
    :description: Full country description (name).

    reverse relations:
    :companies: All the companies in this country.

    """
    code = models.CharField(
        verbose_name=_('Code'),
        max_length=2,
    )

    description = models.CharField(
        verbose_name=_('Description'),
        max_length=128,
    )

    def __unicode__(self):
        return self.description


class EPNManager(models.Manager):
    def get_queryset(self):
        return super(EPNManager, self).get_queryset().select_related(
            'company', 'ipn', 'cpn')


class EPN(models.Model):
    """
    External Part Number.

    :company: The company this EPN is belongs to.
    :cpn: Cosign Part Number.
    :description: The part description
    :epn: The actual part number as string.
    :ipn: Internal Part Number.

    reverse relations:
    :cbomitems: All the CBOM items, that have this EPN.

    """
    company = models.ForeignKey(
        Company,
        verbose_name=_('Company'),
        related_name='epns',
    )

    cpn = models.ForeignKey(
        'aps_bom.IPN',
        verbose_name=_('CPN'),
        related_name='cpns',
        blank=True, null=True,
    )

    description = models.CharField(
        verbose_name=_('Description'),
        max_length=128,
    )

    epn = models.CharField(
        verbose_name=_('EPN'),
        max_length=50,
    )

    ipn = models.ForeignKey(
        'aps_bom.IPN',
        verbose_name=_('IPN'),
        related_name='ipns',
        blank=True, null=True,
    )

    # caching fields only
    company_description = models.CharField(max_length=256, blank=True)
    ipn_code = models.CharField(max_length=256, blank=True)

    objects = EPNManager()

    def __unicode__(self):
        return u'EPN "{0}" - {1}'.format(
            self.epn, self.description)

    def save(self, *args, **kwargs):
        if self.ipn_id is not None:
            self.ipn_code = self.ipn.code
        else:
            self.ipn_code = ''
        self.company_description = self.company.description
        return super(EPN, self).save(*args, **kwargs)

    class Meta:
        ordering = ('epn', )


class IPNManager(models.Model):
    def get_queryset(self):
        return super(IPNManager, self).get_queryset().select_related(
            'price_group', 'shape')


class IPN(models.Model):
    """
    Internal Part Number.

    :code: The actual part number.
    :code2: TODO: Describe this
    :name: A name for this IPN.
    :price_group: The price group, this IPN is in.
    :shape: TODO

    reverse relations:
    :additional_texts: The list of AdditionalText objects attached.
    :boms: The BOM objects this is part of.
    :cpns: The EPN objects this was assigned as a CPN.
    :ipns: The EPN objects this was assigned as an IPN.
    :price_markers: All the price markers of this IPN.

    """
    code = models.CharField(
        verbose_name=_('Code'),
        max_length=50,
    )

    code2 = models.CharField(
        verbose_name=_('Code 2'),
        max_length=50,
        blank=True, null=True,
    )

    name = models.CharField(
        verbose_name=_('Name'),
        max_length=128,
    )

    price_group = models.ForeignKey(
        'aps_bom.PriceGroup',
        verbose_name=_('Price group'),
        related_name='ipns',
    )

    shape = models.ForeignKey(
        'aps_bom.Shape',
        verbose_name=_('Shape'),
        related_name='ipns',
    )

    objects = IPNManager()

    @property
    def price_max(self):
        try:
            latest_marker = PriceMarker.objects.filter(ipn=self).latest('date')
        except PriceMarker.DoesNotExist:
            return
        return latest_marker.price * self.price_group.rate + \
            self.price_group.add

    def __unicode__(self):
        return u'{0} - IPN: {1}'.format(self.name, self.code)


class PriceGroup(models.Model):
    """
    The price group of a part.

    :add: The fixed price addition, that applies to this price group.
    :code: A single character representing the price group.
    :rate: A decimal representing the rate to multiply with the
      IPN.price_marker. TODO

    reverse relations:
    :ipns: All the IPNs that have this price group.

    """
    add = models.DecimalField(
        verbose_name=_('Add'),
        max_digits=10,
        decimal_places=5,
        default=Decimal('0.00000'),
    )

    code = models.CharField(
        verbose_name=_('Code'),
        max_length=1,
    )

    rate = models.DecimalField(
        verbose_name=_('Rate'),
        max_digits=10,
        decimal_places=5,
        default=Decimal('1.00000'),
    )

    def __unicode__(self):
        return u'{0} = Price x {1} + {2}'.format(
            self.code, self.rate, self.add)


class PriceMarker(models.Model):
    """
    TODO

    :ipn: The IPN this price marker belongs to.
    :price: The price this marker defines.
    :date: The date when/from/until this is valid. TODO

    """
    ipn = models.ForeignKey(
        IPN,
        verbose_name=_('Ipn'),
        related_name='price_markers',
    )

    price = models.DecimalField(
        verbose_name=_('Price'),
        max_digits=10,
        decimal_places=5,
    )

    date = models.DateField(
        verbose_name=_('Date'),
    )

    def __unicode__(self):
        return u'{0} - IPN: {1} - Date: {2}'.format(
            self.price, self.ipn.code, self.date)


class Shape(models.Model):
    """
    TODO

    :code: The code of this shape.
    :name: The name/description of this shape.

    reverse relations:
    :ipns: All the IPNs that have this shape.

    """

    code = models.CharField(
        verbose_name=_('Code'),
        max_length=10,
    )

    name = models.CharField(
        verbose_name=_('Name'),
        max_length=200,
    )

    def __unicode__(self):
        return u'Shape: {0}'.format(self.name)


class Unit(models.Model):
    """
    The unit a qty can be in. (see e.g. BOMItem)

    :code: The code of this unit.
    :description: The descriptive name of this unit.

    related:
    :cbomitems: All the CBOMItems, that have this unit.
    :bomitems: All the BOMItems, that have this unit.

    """
    code = models.CharField(
        verbose_name=_('Code'),
        max_length=5,
    )

    description = models.CharField(
        verbose_name=_('Description'),
        max_length=20,
    )

    def __unicode__(self):
        return u'Unit: {0} ({1})'.format(self.code, self.description)



"""Views for the aps_bom app."""
from django.contrib.auth.decorators import login_required
from django.core.urlresolvers import reverse
from django.http import Http404, HttpResponse
from django.views.generic import CreateView, TemplateView, View
from django.utils.decorators import method_decorator

from . import settings
from . import models
from .forms import BOMUploadForm, CBOMUploadForm


class DropDownChoicesAJAXView(View):
    """
    A view for returning drop down choices from certain criteria.

    """
    @method_decorator(login_required)
    def dispatch(self, request, *args, **kwargs):
        action = self.request.GET.get('action')
        obj_id = self.request.GET.get('obj_id')
        if not request.is_ajax() or not hasattr(self, action):
            raise Http404
        return getattr(self, action)(obj_id)

    def get_epn_choices_for_customer(self, obj_id):
        choice_tempalte = '<option value="{value}">{label}</option>'
        choices = '<option value>---------</option>'
        for epn in models.EPN.objects.filter(company_id=obj_id):
            choices += choice_tempalte.format(value=epn.id, label=unicode(epn))
        return HttpResponse(choices)


class BOMUploadView(CreateView):
    """
    Lets a user upload a BOM.csv file and from it create a new BOM instance.

    """
    form_class = BOMUploadForm
    template_name = 'aps_bom/bom_upload.html'

    @method_decorator(login_required)
    def dispatch(self, request, *args, **kwargs):
        return super(BOMUploadView, self).dispatch(request, *args, **kwargs)

    def get_form_kwargs(self):
        kwargs = super(BOMUploadView, self).get_form_kwargs()
        kwargs.update({'files': self.request.FILES})
        return kwargs

    def get_success_url(self):
        return settings.BOM_UPLOAD_SUCCESS_URL


class CBOMDownloadView(TemplateView):
    """
    View that displays a list of cBOMs and allows for downloading the cBOM
    or the BOM, that was generated from it.

    Allows choosing from a list of cBOMs.

    """
    template_name = 'aps_bom/cbom_download.html'

    @method_decorator(login_required)
    def dispatch(self, request, *args, **kwargs):
        self.cbom = None
        if kwargs.get('cbom_pk'):
            try:
                self.cbom = models.CBOM.objects.get(pk=kwargs.get('cbom_pk'))
            except models.CBOM.DoesNotExist:
                raise Http404
        return super(CBOMDownloadView, self).dispatch(request, *args, **kwargs)

    def get_context_data(self, **kwargs):
        ctx = super(CBOMDownloadView, self).get_context_data(**kwargs)
        if self.cbom is not None:
            ctx.update({
                'object': self.cbom,
                'bom': self.cbom.get_bom(),
                'bom_items': self.cbom.get_bom_items(),
                'cbom_file': self.cbom.get_csv_file(),
                'bom_file': self.cbom.get_bom_csv_file(),
            })
        ctx.update({
            'cboms': models.CBOM.objects.all(),
        })
        return ctx


class CBOMUploadView(CreateView):
    """
    Lets a user upload a cBOM.csv file and from it create a new CBOM instance.

    """
    form_class = CBOMUploadForm
    template_name = 'aps_bom/cbom_upload.html'

    @method_decorator(login_required)
    def dispatch(self, request, *args, **kwargs):
        return super(CBOMUploadView, self).dispatch(request, *args, **kwargs)

    def get_form_kwargs(self):
        kwargs = super(CBOMUploadView, self).get_form_kwargs()
        kwargs.update({'files': self.request.FILES})
        return kwargs

    def get_success_url(self):
        return reverse('aps_bom_cbom_download', kwargs={
            'cbom_pk': self.object.pk})



"""URLs for the aps_bom app."""
from django.conf.urls import patterns, url

from . import views


urlpatterns = patterns(
    '',
    url(r'^cbom/upload/$', views.CBOMUploadView.as_view(),
        name='aps_bom_cbom_upload'),
    url(r'^bom/download/(?P<cbom_pk>\d+)/$', views.CBOMDownloadView.as_view(),
        name='aps_bom_cbom_download'),
    url(r'^bom/download/$', views.CBOMDownloadView.as_view(),
        name='aps_bom_cbom_download'),
    url(r'^bom/upload/$', views.BOMUploadView.as_view(),
        name='aps_bom_bom_upload'),
    url(r'^ddl-ajax/', views.DropDownChoicesAJAXView.as_view(),
        name='aps_bom_ddl_ajax'),
)



"""Tests for the models of the aps_bom app."""
import os
from decimal import Decimal

from django.conf import settings
from django.test import TestCase

from ..models import BOM, BOMItem
from .factories import (
    AdditionalTextFactory,
    BOMFactory,
    BOMItemFactory,
    CBOMFactory,
    CBOMItemFactory,
    CompanyFactory,
    CountryFactory,
    EPNFactory,
    IPNFactory,
    PriceGroupFactory,
    PriceMarkerFactory,
    ShapeFactory,
    UnitFactory,
)


class AdditionalTextTestCase(TestCase):
    """Tests for the ``AdditionalText`` model class."""
    longMessage = True

    def test_instantiation(self):
        """Test instantiation of the ``AdditionalText`` model."""
        additionaltext = AdditionalTextFactory()
        self.assertTrue(additionaltext.pk)


class BOMTestCase(TestCase):
    """Tests for the ``BOM`` model class."""
    longMessage = True

    def test_instantiation(self):
        """Test instantiation of the ``BOM`` model."""
        bom = BOMFactory()
        self.assertTrue(bom.pk)


class BOMItemTestCase(TestCase):
    """Tests for the ``BOMItem`` model class."""
    longMessage = True

    def test_instantiation(self):
        """Test instantiation of the ``BOMItem`` model."""
        bomitem = BOMItemFactory()
        self.assertTrue(bomitem.pk)


class CBOMTestCase(TestCase):
    """Tests for the ``CBOM`` model class."""
    longMessage = True

    def setUp(self):
        self.cbom = CBOMFactory()

    def test_instantiation(self):
        """Test instantiation of the ``CBOM`` model."""
        self.assertTrue(self.cbom.pk)

    def test_get_bom(self):
        self.cbomitem = CBOMItemFactory(bom=self.cbom, consign=False)
        self.assertEqual(
            self.cbom.get_bom(),
            BOM(description=self.cbom.description, ipn=None))

    def test_get_bom_csv_file(self):
        path = os.path.join(settings.MEDIA_URL, 'BOM_{0}.csv'.format(
            self.cbom.id))
        self.cbomitem = CBOMItemFactory(bom=self.cbom, consign=False)
        self.assertEqual(self.cbom.get_bom_csv_file(), path)

    def test_get_bom_items(self):
        self.cbomitem = CBOMItemFactory(bom=self.cbom, consign=False)
        CBOMItemFactory(bom=self.cbom, consign=True)
        self.assertEqual(
                self.cbom.get_bom_items()[0],
                BOMItem(ipn=self.cbomitem.epn.ipn,
                        position=self.cbomitem.position,
                        qty=self.cbomitem.qty, unit=self.cbomitem.unit))

    def test_get_csv_file(self):
        path = os.path.join(settings.MEDIA_URL, 'cBOM_{0}.csv'.format(
            self.cbom.id))
        self.cbomitem = CBOMItemFactory(bom=self.cbom, consign=False)
        self.assertEqual(self.cbom.get_csv_file(), path)


class CBOMItemTestCase(TestCase):
    """Tests for the ``CBOMItem`` model class."""
    longMessage = True

    def test_instantiation(self):
        """Test instantiation of the ``CBOMItem`` model."""
        cbomitem = CBOMItemFactory()
        self.assertTrue(cbomitem.pk)


class CompanyTestCase(TestCase):
    """Tests for the ``Company`` model class."""
    longMessage = True

    def test_instantiation(self):
        """Test instantiation of the ``Company`` model."""
        company = CompanyFactory()
        self.assertTrue(company.pk)


class CountryTestCase(TestCase):
    """Tests for the ``Country`` model class."""
    longMessage = True

    def test_instantiation(self):
        """Test instantiation of the ``Country`` model."""
        country = CountryFactory()
        self.assertTrue(country.pk)


class EPNTestCase(TestCase):
    """Tests for the ``EPN`` model class."""
    longMessage = True

    def test_instantiation(self):
        """Test instantiation of the ``EPN`` model."""
        epn = EPNFactory()
        self.assertTrue(epn.pk)


class IPNTestCase(TestCase):
    """Tests for the ``IPN`` model class."""
    longMessage = True

    def test_instantiation(self):
        """Test instantiation of the ``IPN`` model."""
        ipn = IPNFactory()
        self.assertTrue(ipn.pk)

    def test_price_max(self):
        """Test for the ``price_max`` property."""
        ipn = IPNFactory()
        self.assertIsNone(ipn.price_max, msg=(
            'Without a PriceMarker, the price_max should be None.'))
        PriceMarkerFactory(ipn=ipn)
        self.assertEqual(ipn.price_max, Decimal('54.32100000'), msg=(
            'The price max was not calculated correctly.'))


class PriceGroupTestCase(TestCase):
    """Tests for the ``PriceGroup`` model class."""
    longMessage = True

    def test_instantiation(self):
        """Test instantiation of the ``PriceGroup`` model."""
        pricegroup = PriceGroupFactory()
        self.assertTrue(pricegroup.pk)


class PriceMarkerTestCase(TestCase):
    """Tests for the ``PriceMarker`` model class."""
    longMessage = True

    def test_instantiation(self):
        """Test instantiation of the ``PriceMarker`` model."""
        pricemarker = PriceMarkerFactory()
        self.assertTrue(pricemarker.pk)


class ShapeTestCase(TestCase):
    """Tests for the ``Shape`` model class."""
    longMessage = True

    def test_instantiation(self):
        """Test instantiation of the ``Shape`` model."""
        shape = ShapeFactory()
        self.assertTrue(shape.pk)


class UnitTestCase(TestCase):
    """Tests for the ``Unit`` model class."""
    longMessage = True

    def test_instantiation(self):
        """Test instantiation of the ``Unit`` model."""
        unit = UnitFactory()
        self.assertTrue(unit.pk)



"""Settings that need to be set in order to run the tests."""
import os

DEBUG = True

SITE_ID = 1

APP_ROOT = os.path.abspath(
    os.path.join(os.path.dirname(__file__), '..'))

LOGIN_URL = '/admin/'

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': ':memory:',
    }
}

ROOT_URLCONF = 'aps_bom.tests.urls'

STATIC_URL = '/static/'
MEDIA_URL = '/media/'
STATIC_ROOT = os.path.join(APP_ROOT, '../app_static')
MEDIA_ROOT = os.path.join(APP_ROOT, '../app_media')
STATICFILES_DIRS = (
    os.path.join(APP_ROOT, 'static'),
)

TEMPLATE_DIRS = (
    os.path.join(APP_ROOT, 'tests/test_app/templates'),
)

COVERAGE_REPORT_HTML_OUTPUT_DIR = os.path.join(
    os.path.join(APP_ROOT, 'tests/coverage'))
COVERAGE_MODULE_EXCLUDES = [
    'tests$', 'settings$', 'urls$', 'locale$',
    'migrations', 'fixtures', 'admin$', 'django_extensions',
]

EXTERNAL_APPS = [
    'django.contrib.admin',
    'django.contrib.admindocs',
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.messages',
    'django.contrib.sessions',
    'django.contrib.staticfiles',
    'django.contrib.sitemaps',
    'django.contrib.sites',
    'django_jasmine',
    'django_nose',
    'django_libs',
]

INTERNAL_APPS = [
    'aps_bom',
    'aps_bom.tests.test_app',
]

INSTALLED_APPS = EXTERNAL_APPS + INTERNAL_APPS
COVERAGE_MODULE_EXCLUDES += EXTERNAL_APPS

SECRET_KEY = 'foobar'



"""Settings for the aps_bom app."""
from django.conf import settings
from django.core.urlresolvers import reverse_lazy

# default for the headlines in the csv file
BOM_CSV_FIELDNAMES = getattr(settings, 'APS_BOM_BOM_CSV_FIELDNAMES',
                             ["Position", "IPN", "Description", "QTY", "Unit",
                              "Shape"])

CBOM_CSV_FIELDNAMES = getattr(settings, 'APS_BOM_BOM_CSV_FIELDNAMES',
                              ["Position", "EPN", "Description", "QTY", "Unit",
                               "consign"])

BOM_UPLOAD_SUCCESS_URL = getattr(settings,
                                 'APS_BOM_BOM_UPLOAD_VIEW_SUCCESS_URL',
                                 reverse_lazy('aps_bom_bom_upload'))



"""URLs to run the tests."""
from django.conf import settings
from django.conf.urls import patterns, include, url
from django.conf.urls.static import static
from django.contrib import admin
from django.contrib.staticfiles.urls import staticfiles_urlpatterns


admin.autodiscover()


urlpatterns = static(settings.MEDIA_URL, document_root=settings.MEDIA_ROOT)
urlpatterns += staticfiles_urlpatterns()

urlpatterns += patterns(
    '',
    url(r'', include('aps_bom.urls')),
    url(r'^admin/', include(admin.site.urls)),
)
